#include <jni.h>
#include <string>
#include "Ceda/Core/cxUtils/Tracer.h"
#include "Ceda/Core/cxLss/ILogStructuredStore.h"
#include "Ceda/Core/cxLss/LssSettings.h"
#include "Ceda/Core/cxOperation/UTRoot.h"
#include "Ceda/Core/cxOperation/IWorkingSetMachine.h"
#include "Ceda/Core/cxWorkingSetIpc/MultiWorkingSetIPC.h"
#include "Ceda/Core/cxPersistStore/IPersistStore.h"
#include "Ceda/Core/cxObject/IObjectVisitor.h"
#include "Ceda/Core/cxObject/PrintReflectedVariable.h"
#include "Ceda/Core/cxUtils/HPTime.h"
#include "Ceda/Core/cxUtils/CedaAssert.h"
#include "Acme/Pizza/PizzaDeliveries.h"
#include "Acme/Pizza/Shapes.h"
#include "Acme/Pizza/DateTime.h"

using namespace ceda;

extern "C" void Register_Project_Pizza();

void TwoWorkingSets()
{
    Tracer() << "TwoWorkingSets example\n";
    ceda::TraceIndenter indent;

    const ceda::xstring UtEntryName("MyApp");
    ceda::xstring path = "/sdcard/temp/cedatest";
    ceda::PersistStore* pstore = ceda::OpenPersistStore(path.c_str(), ceda::OM_CREATE_ALWAYS);
    {
        // Open or create a PSpace
        ceda::WPSpace pspace(ceda::OpenPSpace(pstore, "MyPSpace"));

        // Open or create two working sets in the PSpace
        ceda::WorkingSetMachine* ws1 = nullptr;
        ceda::WorkingSetMachine* ws2 = nullptr;
        {
            ceda::PSpaceTxn txn;
            ws1 = ceda::OpenWorkingSetMachine("WS1",true);
            ws2 = ceda::OpenWorkingSetMachine("WS2",false);
        }

        {
            ceda::Iocp* iocp = ceda::CreateIocp();

            // Create server and client
            ceda::MServer* server;
            ceda::MClient* client;
            {
                const ceda::WsipcSessionProtocolId protocolId = { "TwoWorkingSets", 1 };
                ceda::WsipcHostInfo localHostInfo;
                ceda::PSpaceTxn txn;
                server = CreateSimpleWsipcServer(iocp, ws1, protocolId, 3000);
                client = CreateSimpleWsipcClient(iocp, ws2, protocolId, "127.0.0.1", 3000);
            }

            // Bootstrap TestDS under the UT root
            {
                ceda::CSpaceTxn txn;
                GetUTRoot(ws1).Map[UtEntryName].insert(0,ceda::DollarNew(new pizza::TPizzaOrder));
            }

            Tracer() << "Showing result on ws1\n";
            {
                ceda::PSpaceTxn txn;
                pizza::TPizzaOrder* t = ceda::GetUTRootEntry<pizza::TPizzaOrder>(ws1,UtEntryName);
                ceda::PrintInstance(Tracer(), t);
            }

            // Generate some operations
            for (int i=0 ; i < 10 ; ++i)
            {
                ceda::CSpaceTxn txn;
                pizza::TPizzaOrder* t = ceda::GetUTRootEntry<pizza::TPizzaOrder>(ws1,UtEntryName);
                t->Id = i;             // Generate assignment operation
            }

            Tracer() << "Showing result on ws1\n";
            {
                ceda::PSpaceTxn txn;
                pizza::TPizzaOrder* t = ceda::GetUTRootEntry<pizza::TPizzaOrder>(ws1,UtEntryName);
                ceda::PrintInstance(Tracer(), t);
            }

            // Poll for quiescence
            Tracer() << "Polling for quiescence\n";
            while(1)
            {
                Sleep(1000);
                ceda::PSpaceTxn txn;
                Tracer() << "HistoryBufferSize(ws2) = " << HistoryBufferSize(ws2) << '\n';
                if (HistoryBufferSize(ws2) == 11) break;
            }

            Tracer() << "Showing result on ws2\n";
            {
                ceda::PSpaceTxn txn;
                pizza::TPizzaOrder* t = ceda::GetUTRootEntry<pizza::TPizzaOrder>(ws2,UtEntryName);
                ceda::PrintInstance(Tracer(), t);
            }

            Close(server);
            Tracer() << "closed server\n";
            Close(client);
            Tracer() << "closed client\n";

            Close(iocp);
            Tracer() << "closed iocp\n";
        }

        Close(ws1);
        Close(ws2);
        Tracer() << "closed working sets\n";
    }
    Close(pstore);
    Tracer() << "closed pstore\n";
}




extern "C"
JNIEXPORT jstring

JNICALL
Java_com_cedanet_cedaandroidexample_MainActivity_stringFromJNI(
        JNIEnv *env,
        jobject /* this */) {

    ceda::SetTraceFile("stdout",false,false);

    ceda::StringStream os;
    os << "hello\n";

    try
    {
        Register_Project_Pizza();
        TwoWorkingSets();
    }
    catch(ceda::IException& e)
    {
        os << e << '\n';
    }

    /*
    try
    {
        bool createdNew;
        LssSettings settings;
        ILogStructuredStore* lss = CreateOrOpenLSS("/sdcard/temp/cedatest", NULL, createdNew,
                                                   EOpenMode::OM_OPEN_ALWAYS, settings);
        lss->Close();
    }
    catch(ceda::IException& e)
    {
        os << e << '\n';
    }
     */

    return env->NewStringUTF(os.GetString().c_str());
}
